﻿//-------------------------------------------------------------------------
// MAIN 
// NOTES: THIS FILE SHOULD REMAIN UNMODIFIED BY NON-ENGINE  CODE.  IT IS
// THE MAIN ENTRY POINT FOR THE APPLICATION.  SEE 'GAMEMAIN' FOR APPLICATION
// LEVEL CODE AND ROUTINES.
//-------------------------------------------------------------------------
"use strict";

//-------[ WINDOW CALLBACKS ]----------------------------------------------
window.addEventListener('load', main_init, false);
//-------------------------------------------------------------------------

//-------[ GLOBAL VARS ]---------------------------------------------------
var DEBUG = true; //debug mode (bool)
var GAME_MODE = 0; // webserver | webkit | winjs
var GAME_ENV = 0; // firefox | chrome | safari | edge | winjs
var USE_THREADING = false; //set in main by browser detect
var AACOLLISION_MANAGER = null; //var ref for later
var COLLISION_BINS_X = 15; var COLLISION_BINS_Y = 9;
var GAME_STATE = 0; //unloaded, uninitialized, always starts at 0
//-------------------------------------------------------------------------
// creation of global COLLISION_MANAGER -- initialization called from
// main application init
//-------------------------------------------------------------------------


//-------[ GLOBAL DETECT FUNCTION ]----------------------------------------
GAME_MODE = "webserver"; // webserver | webkit | winjs
var WinJS = window.parent.WinJS || null;
if (WinJS != null) { GAME_MODE = "winjs"; } else {
    var fs = 0; try { fs = require('fs'); } catch (e) {  }
    if (fs) { GAME_MODE = "webkit"; }
}
//------------------------------------------------------------------------- 


//-------[ GLOBAL BROWSER DETECT FUNCTION ]--------------------------------
navigator.sayswho = (function () {  /* courtesy of kennebec stackOverflow*/
    var ua = navigator.userAgent, tem,
    M = ua.match(/(opera|chrome|safari|firefox|msie|trident(?=\/))\/?\s*(\d+)/i) || [];
    if (/trident/i.test(M[1])) {
        tem = /\brv[ :]+(\d+)/g.exec(ua) || [];
        return 'IE ' + (tem[1] || '');
    }
    if (M[1] === 'Chrome') {
        tem = ua.match(/\b(OPR|Edge)\/(\d+)/);
        if (tem != null) return tem.slice(1).join(' ').replace('OPR', 'Opera');
    }
    M = M[2] ? [M[1], M[2]] : [navigator.appName, navigator.appVersion, '-?'];
    if ((tem = ua.match(/version\/(\d+)/i)) != null) M.splice(1, 1, tem[1]);
    return M.join(' ');
})();  GAME_ENV = navigator.sayswho;
console.log("AAENGINE: GAME_ENV detected as: " + GAME_ENV);
//-------------------------------------------------------------------------

//-------[ MAIN INIT ]-----------------------------------------------------
/*  
    PURPOSE: main_init is called in response to the onload function of the
    application window.  Should make all necessary global handlers, 
    set all needed configuration variables, gets everything ready,
    and then kicks off main_loop() as the main applicaiton loop.
*/
//-------------------------------------------------------------------------
function main_init() {
    console.log("AAENGINE: CORE START -----------------------------------");

    if (GAME_ENV.indexOf('Chrome') != -1) {

        USE_THREADING = true;
    }
    //NOTE: Initialization routines and loading sequence ORDER IS IMPORTANT.
    
    //-----[ INITIALIZE GAME WINDOW ]--------------------------------------
    GAME_WINDOW.init(16, 9 , "horizontal");

    //-----[ INITIALIZE DEBUG TOOLS IF DEBUGGING ]-------------------------
    if (DEBUG) {
        FPS_COUNTER.init();
        DEBUG_DISPLAY.init();
    }

    //-----[ INITIALIZE 3D MANAGER ]---------------------------------------
    AA3D_MANAGER.init();

    //-----[ INITIALIZE 2D DEBUG OVERLAY ]---------------------------------
    if (DEBUG) { DEBUG_DISPLAY.initDebugOverlay(); }

    //-----[ INITIALIZE MODEL MANAGER ]------------------------------------
    AAMODEL_MANAGER.init();

    //-----[ INITIALIZE COLLISION MANAGER ]--------------------------------
    if (!USE_THREADING) {
        AACOLLISION_MANAGER = new CollisionManager();
        AACOLLISION_MANAGER.init(COLLISION_BINS_X, COLLISION_BINS_Y);
    }

    //-----[ START ASSOCIATED THREADS ]------------------------------------
    if (USE_THREADING) {
        THREAD_MANAGER.init();
        THREAD_MANAGER.createThread(1); //1 = thread ID | game object mgr
        //THREAD_MANAGER.createThread(2); //1 = thread ID | second thread
    }

    //-----[ INITIALIZE INPUT MANAGER ]------------------------------------
    INPUT_MANAGER.init();

    //-----[ START APPLICATION ]-------------------------------------------
    window.focus();
    //TODO: Use real loading locks at startup instead of hacking around race condition :)
    setTimeout("requestAnimationFrame(main_application_loop)", 1000);
    
}
//-------------------------------------------------------------------------

function main_application_loading_callback() {

    //-----[ INITIALIZE THE GAME ]-----------------------------------------
    GAME_WINDOW.resize(); //throw a resize - needed by certain platforms
    gameInit();

    if (!USE_THREADING) {
        GAME_STATE = 2; //proceed to run level 2
    } else {
        GAME_STATE = 1; //go to run level 1 and wait for thread callback
    }
}

function main_applicaiton_threading_callback() {
    
    if (GAME_STATE == 1) {
        console.log("AAENGINE: THREADING ENABLED AND ESTABLISHED - RUN LEVEL 2");
        GAME_STATE = 2;
    } else {
        console.log("AAENGINE: THREADING ENABLED - WAITING ON ASSETS - RUN LEVEL 1");
        setTimeout('main_application_threading_callback()', 30);
    }
}

//-------[ MAIN APPLICATION LOOP ]-----------------------------------------
/*  
    PURPOSE: main_application_loop is called back from requestAnimationFrame
    and is the main loop of the entire application (ie calls update on all
    other subsystems in a specified order), and then requests the next
    frame.
*/
//-------------------------------------------------------------------------
function main_application_loop(curTime) {
    //if we are in state 2, things are now loaded but not yet linked
    if (GAME_STATE == 2) {
        if (USE_THREADING) {
            THREAD_MANAGER.linkModelsToObjects();
            GAME_STATE = 3;
        } else {
            AAMODEL_MANAGER.linkMeshesToGameObjects();
        }
        
    }

    //we have now linked models and meshes, and done our cloning, so we're ready
    if (GAME_STATE == 4) {
        //-----[ RED SLICE : INPUT COLLECTION ]--------------------------------
        if (DEBUG) { FPS_COUNTER.setTimeSlice('rgba(255,0,0,0.5)'); } //RED

        //update input - should always be done first per-frame
        INPUT_MANAGER.update();

        //-----[ ORANGE SLICE (END RED) : DEBUG DISPLAY UPDATE ]---------------
        if (DEBUG) { FPS_COUNTER.setTimeSlice('rgba(255,200,0,0.5)'); }
        if (DEBUG) { DEBUG_DISPLAY.update(); }

        //-----[ GREEN SLICE (END ORANGE) : THREAD MANAGER UPDATE ]------------
        if (DEBUG) { FPS_COUNTER.setTimeSlice('rgba(0,255,0,0.5)'); }

        //update thread manager, or update all game objects (threaded game
        //objects update themselves asynchronously)
        if (USE_THREADING) {
            THREAD_MANAGER.update();
        } else {
            AACOLLISION_MANAGER.preupdate();
            AA3D_MANAGER.updateAllGameObjects();
        }

        //update all models based on what happened above
        AAMODEL_MANAGER.update();

        //-----[ BLUE SLICE (END GREEN) : COLLISION DETECTION ]----------------- 
        if (DEBUG) { FPS_COUNTER.setTimeSlice('rgba(0,0,255,0.5)'); }
        if (!USE_THREADING) { AACOLLISION_MANAGER.update(); }

        //-----[ PURPLE SLICE (END BLUE) : 3D RENDERING ]----------------------- 
        if (DEBUG) { FPS_COUNTER.setTimeSlice('rgba(128,0,255,0.5)'); }

        //render the 3D Scene
        AA3D_MANAGER.render();

        //update time this frame took 
        if (DEBUG) { FPS_COUNTER.update(curTime); }
    }

    //should be called last - get the next frame
    requestAnimationFrame(main_application_loop);
}
//-------------------------------------------------------------------------


//-------[ MAIN EXIT ]-----------------------------------------------------
/* 
    Purpose - called to leave the application and close
    down any and all handles.  The end. 
*/
//--------------------------------------------------------------------------
function main_exit() {
    
    console.log("AAENGINE: Application Close");
    //if we are threading, mark active threads as inactive so updates stop
    if (USE_THREADING) {
        THREAD_MANAGER.stop();
    }

    //attempt to close the WinJS way
    if (GAME_MODE == 'winjs') {
        var app = window.parent.WinJS.Application;
        app.stop();
        window.parent.window.close();
       
    }
    else { //attempt to close this tab the node-webkit way
        try {
            // Load native UI library
            var gui = require('nw.gui');
            // Get the current window
            var win = gui.Window.get();
            win.close(true);
        }
        catch (e) {
            //silent fail means not a node_webkit instance
        }
        //try to close the browser-y way
        try {
            window.close();
        }
        catch (e) {
            //fail here means browser does not allow close
        }
        //go to the studios page if all else fails
        window.location.replace("http://magic.rit.edu/studios/");
    }
}
//-------------------------------------------------------------------------