﻿"use strict";

class AAModelManager {

    //-----[ AA3D MANAGER: CONSTRUCTOR ]-----------------------------------
    constructor() {

        //pattern to ensure singleton of GAME_WINDOW object
        //there can be only one...
        if (!AAMODEL_MANAGER) {
            AAMODEL_MANAGER = this;

        } else {
            console.log("AAENGINE: ERROR: Attempt to create multiple model managers." +
                        "\nAAENGINE: ERROR: Only first instanciation will be used.");
            return AAMODEL_MANAGER;
        }
    }

    //---------------------------------------------------------------------
    //init.  for now this just creates a bunch of dummy objects and loads them
    //---------------------------------------------------------------------
    init() {

        //create a game_objects array to store game objects
        this._gameObjectsArray = []; //array of all game objects in scene
        this._gameModelsMasterList = []; //the base models to copy from

        for (var i = 0; i < TOTAL_NUM_GAME_OBJECTS; i++) {
            this._gameObjectsArray[i] = [];
        }

        //------------------------------------------------------------------
        //load all models
        //------------------------------------------------------------------
        this._assetsLoaded = 0;
        this._total3Dassets = 0;
        this._loadingManager = new THREE.LoadingManager();

        //simple cube is the default geometry for position 0
        var g = new THREE.CubeGeometry(1, 1, 1, 1, 1, 1);
        var m = new THREE.MeshLambertMaterial(0xffffff);
        this._gameModelsMasterList[0] = new THREE.Mesh(g, m);

        //simple sphere is the defult geometry for position 1
        g = new THREE.SphereGeometry(0.5, 12, 12, 0, Math.PI*2, 0, Math.PI);
        m = new THREE.MeshLambertMaterial(0xffffff);
        this._gameModelsMasterList[1] = new THREE.Mesh(g, m);

        //------------------------------------------------------------------
        //ITEM LOADING PROGRESS CALLBACK
        //------------------------------------------------------------------
        this._loadingManager.onProgress = function (item, loaded, total) {
            console.log("AAENGINE: Loaded item " + item +
            " as #" + loaded + " of " + total + " 3D assets.");

            this._assetsLoaded++;
            this._total3Dassets = total;

            //TODO: Stuff with a load bar...
        }
        //------------------------------------------------------------------
        
        
        //load the model
        var loader = new THREE.JSONLoader(this._loadingManager);
        var url = '';

        //load and parse the model information file
        var mmj = 0; //model list information from jason
        loadJSON(function (response) {
            // Parse JSON string into object
            mmj = JSON.parse(response);

            for (var m = 0; m < mmj.models.length; m++) {
                console.log(mmj.models[m].name);
                url = mmj.models[m].path;
                name = mmj.models[m].name;
                loader.load(url, modelLoadingCallback(url, name));
            }

        }, './models/model_list.json');

        //load the textures
        //var textureLoader = new THREE.TextureLoader(this._loadingManager);
        //var normTexture = textureLoader.load("./models/test_ship/normalmap.png", textureLoadingCallback());
        //var occlTexture = textureLoader.load("./models/test_ship/texmap.png", textureLoadingCallback());

        //------------------------------------------------------------------
        //ITEM LOADING FINAL CALLBACK
        //------------------------------------------------------------------
        this._loadingManager.onLoad = function (item, loaded, total) {
            console.log("AAENGINE: All Assets Loaded " + this._total3Dassets);


            for (var m = 0; m < mmj.models.length; m++) {
                //get handles to all current model
                AAMODEL_MANAGER._gameModelsMasterList[mmj.models[m].index] =
                    AA3D_MANAGER.SCENE.getObjectByName(mmj.models[m].name);
                //remove the master version from the scene 
                //(clones will be used later)
                AA3D_MANAGER.SCENE.remove(
                    AAMODEL_MANAGER._gameModelsMasterList[
                    mmj.models[m].index]);
            }

            //callback to the main application thread that things are loaded
            //-------------------------------------------------------------
            main_application_loading_callback();
            //-------------------------------------------------------------
        }
    }

    linkMeshesToGameObjects(gos) {
        if (USE_THREADING) {
            for (var i = 0; i < gos.length; i++) {
                
                var id = gos[i][0]; //
                var mesh = AAMODEL_MANAGER._gameModelsMasterList[gos[i][1]].clone();
                AA3D_MANAGER.SCENE.add(mesh);

                var p = new Float32Array(3);
                var r = new Float32Array(3);
                var s = new Float32Array(3); s[0] = 1; s[1] = 1; s[2] = 1;
                AAMODEL_MANAGER.loadModel(id, mesh, p, r, s, 99, [0, 0, 0, 0, 0, 0, 0, 0]);
            }
            GAME_STATE = 4; //done linking, so move to run state 4
        } else {
            for (var i = 0; i < AA3D_MANAGER._gameObjectsArray.length; i++) {
                //gonna hope these are already loaded?
                var id = AA3D_MANAGER._gameObjectsArray[i]._id;
                var mesh = this._gameModelsMasterList[
                    AA3D_MANAGER._gameObjectsArray[i]._meshID].clone();
                AA3D_MANAGER.SCENE.add(mesh);

                var p = new THREE.Vector3(0, 0, 0);
                var r = new THREE.Vector3(0, 0, 0);
                var s = new THREE.Vector3(1, 1, 1);
                AAMODEL_MANAGER.loadModel(id, mesh, p, r, s, 99, [0, 0, 0, 0, 0, 0, 0, 0]);
            }
            GAME_STATE = 4; //done linking, so move to run state 4
            console.log("GAME STATE NOW " + GAME_STATE);
        }
    }

    //---------------------------------------------------------------------
    //---------------------------------------------------------------------
    //update all the game objects in the scene based on their model records
    //---------------------------------------------------------------------
    update() {

        for (var i = 0; i < this._gameObjectsArray.length; i++) {
            
            var m = this._gameObjectsArray[i];
            if (m[0] != 200) {
                
                //handle to mesh at this meshID
                var mesh = m[1]; //this._gameObjectModels[m[1]];

                if (USE_THREADING) {
                    //position this model for rendering, as it has a valid id
                    mesh.position.set(m[2][0], m[2][1], m[2][2]);
                    mesh.rotation.set(m[3][0], m[3][1], m[3][2]);
                    mesh.scale.set(m[4][0], m[4][1], m[4][2]);

                } else {
                    //position this model for rendering, as it has a valid id
                    mesh.position.set(m[2].x, m[2].y, m[2].z);
                    mesh.rotation.set(m[3].x, m[3].y, m[3].z);
                    mesh.scale.set(m[4].x, m[4].y, m[4].z);
                }
                //render the model in the 2D debug overlay if debugging
                if (DEBUG && USE_THREADING) {
                    if (DEBUG_DISPLAY._2dDebugVisibility) {
                        var lineColor = 'rgba(255,255,255,1.0)';
                        var fillColor = 'rgba(255,255,255,0.0)';
                        if (m[7]) {
                            lineColor = 'rgba(255,0,0,1.0)';
                            fillColor = 'rgba(255,0,0,0.6)';
                        }
                       //console.log(type);
                        if (m[5] == 2) {
                            DEBUG_DISPLAY.drawDebugPoly(
                            m[6][0], //x
                            m[6][1], //y
                            m[6][2], //points after scale and rotation...
                            GAME_WINDOW.GAME_WINDOW_SCALE,
                            fillColor,
                            lineColor);
                            //fill color, stroke color
                        }
                        else if (m[5] == 1) {
                            DEBUG_DISPLAY.drawDebugCircle(
                            m[6][0],
                            m[6][1],
                            m[6][2] * GAME_WINDOW.GAME_WINDOW_SCALE,
                            fillColor, lineColor);
                            //fill color, stroke color) 
                        }

                        else if (m[5] == 0) {
                            DEBUG_DISPLAY.drawDebugBox(
                            m[6][0], m[6][1], m[6][2], m[6][3],
                            fillColor, lineColor);
                        }
                    }
                }
            }
        }
    }

    //---------------------------------------------------------------------
    //this loads a model by creating a 'model record' - id, mesh,
    //pos, rot, scale.  later, this will link to files and whatnot?
    //---------------------------------------------------------------------
    loadModel(id, mesh, pos, rot, scale, type, data, meshID) {
        //TODO: make this a real loading function later
        this._gameObjectsArray[id] = [id, mesh, pos, rot, scale, type, data, meshID];
    }

    //---------------------------------------------------------------------
    //this updates a model record from a game objects 'render' method, or
    //from the thread managing it (depending on if threading is used).
    //either way, it updates the pos/rot/scale of an object at that id
    //---------------------------------------------------------------------
    updateModel(id, pos, rot, scale, meshID, type, data, collision) {
        
        //index 1 is mesh - don't touch (yet)
        if (USE_THREADING && GAME_STATE >= 4) {
            
            //if _gameObjArraypid][7] != meshID then it changed...
            this._gameObjectsArray[id][7] = meshID;

            this._gameObjectsArray[id][2][0] = pos[0];
            this._gameObjectsArray[id][2][1] = pos[1];
            this._gameObjectsArray[id][2][2] = pos[2];

            this._gameObjectsArray[id][3][0] = rot[0];
            this._gameObjectsArray[id][3][1] = rot[1];
            this._gameObjectsArray[id][3][2] = rot[2];

            this._gameObjectsArray[id][4][0] = scale[0];
            this._gameObjectsArray[id][4][1] = scale[1];
            this._gameObjectsArray[id][4][2] = scale[2];

        } else if (GAME_STATE == 4) {

            this._gameObjectsArray[id][7] = meshID;

            this._gameObjectsArray[id][2] = pos;
            this._gameObjectsArray[id][3] = rot;
            this._gameObjectsArray[id][4] = scale;
        }
        

        if (DEBUG) {
            this._gameObjectsArray[id][5] = type;
            if (type < 4) { //update the collision shape points if sent
                for (var i = 0; i < data.length; i++) {
                    this._gameObjectsArray[id][6][i] = data[i];
                }
            }

            this._gameObjectsArray[id][7] = collision;

        } 
    }
    //---------------------------------------------------------------------
}


//-------------------------------------------------------------------------
// creation of global MODEL_MANAGER -- initialization called from
// main application init
//-------------------------------------------------------------------------
var AAMODEL_MANAGER = AAMODEL_MANAGER || new AAModelManager();
//-------------------------------------------------------------------------


//-------------------------------------------------------------------------
// modelLoadingCallback (called when a model is loaded, must be public)
//-------------------------------------------------------------------------
function modelLoadingCallback(url, name) {
    return function (geometry, materials, metadata) {

        var mat = new THREE.MeshFaceMaterial(materials);
        //var mat = new THREE.MeshLambertMaterial(0xffffff);
        //var mat = new THREE.MeshPhongMaterial(materials);
        //var mat = new THREE.MeshDepthMaterial(materials);
        var handle = new THREE.Mesh(geometry, mat);
        handle.name = name;
        AA3D_MANAGER.SCENE.add(handle);
    }
}
//-------------------------------------------------------------------------


//-------------------------------------------------------------------------
// texture loading callback, called when a texture is loaded.  must be 
// public
//-------------------------------------------------------------------------
function textureLoadingCallback() {
    return function (texture) {
        new THREE.Texture(texture);
    }
}


//-------------------------------------------------------------------------
// loadJSON - helper function that loads JSON from a file specified in
// path, and calls a provided callback (must be anonymous)
//-------------------------------------------------------------------------
function loadJSON(callback, path) {

    var xobj = new XMLHttpRequest();
    xobj.overrideMimeType("application/json");
    xobj.open('GET', path, true); //true = asynchonous  
    xobj.onreadystatechange = function () {
        console.log(xobj.status + " " + xobj.readyState);
        //state check - NOTE: TODO: currently allowing status 0 for debugging
        //should be changed to only be 200 in final implementation builds
        if (xobj.readyState == 4 && (xobj.status == "200" || xobj.status == "0")) {
            // Required use of an anonymous callback as .open will 
            //NOT return a value but simply returns 
            //undefined in asynchronous mode
            callback(xobj.responseText);
        }
    };
    xobj.send(null);
}
//-------------------------------------------------------------------------