﻿"use strict";
//--------------------------------------------------------------------------
// NOTE: Windows / WinJS Currently imposes a hard limit of 25 worker 
// threads for any application (not sure if this includes the 'main' 
// non-worker as well) other browsers have no such limitations

// purpose: Create a Thread Manager that creates and then updates 
// all threads as needed var THREAD_TIMER_0 = 0;
//--------------------------------------------------------------------------
class ThreadManager {

    //----------------------------------------------------------------------
    constructor() {
        //pattern to ensure singleton of GAME_WINDOW object
        //there can be only one...
        if (!THREAD_MANAGER) {
            THREAD_MANAGER = this;
        } else {
            console.log("AAENGINE: ERROR: Attempt to create multiple thread managers." +
                      "\nAAENGINE: ERROR: Only first instanciation will be used.");
            return THREAD_MANAGER;
        }
    }

    //-----[ THREAD_MANAGER: INIT ]-----------------------------------------
    // init the thread manager
    //----------------------------------------------------------------------
    //array of individual workers 
    //TODO: One of these for each KIND of rendered object?  
    //Particle vs.not?
    init() {
        this._threadPool = []; 
        this._threadTimer = 0;
        this._updateCounter = 0;
        this._numGameObjects = 0;
    }
    //---------------------------------------------------------------------


    //-----[ THREAD_MANAGER: UPDATE ]--------------------------------------
    // init the thread manager
    //---------------------------------------------------------------------
    update() {
        // this is an update per frame called on the thread manager.  
        // currently not used for anything...
    }
    //---------------------------------------------------------------------


    //-----[ THREAD_MANAGER: CREATE THREAD ]-------------------------------
    // init the thread manager
    //---------------------------------------------------------------------
    createThread(id) {
        if (window.Worker) {
            //-------------------------------------------------------------
            // if ID is one, thread type is to manage game objects
            //-------------------------------------------------------------
            if (id == 1) {
                this.thread = new Worker(
                    "js/core_engine/worker_threads/ThreadedGameObjectManager.js");

                //---------------------------------------------------------
                // onmessage handler for thread (callback to main UI)
                //---------------------------------------------------------
                this.thread.onmessage = function (e) {
                    //-----------------------------------------------------
                    //callback 0 - normal game object processing
                    //-----------------------------------------------------
                    if (e.data[e.data.length - 1] == 0) {
                        for (var i = 0; i < e.data.length - 5; i++) {
                            
                            //update model manager with information
                            //on the game objects (pos/rot/scale/etc)
                            AAMODEL_MANAGER.updateModel(i,
                                e.data[i][0], e.data[i][1], e.data[i][2],
                                e.data[i][3], e.data[i][4], e.data[i][5],
                                e.data[i][6]);
                        }

                        //figure out timing information 
                        var t = performance.now();
                        var dt = t - self._threadTimer;
                        self._threadTimer = t;

                        //if debugging, draw the bins
                        if (DEBUG) {
                            if (e.data[e.data.length - 5] != 0) {
                                DEBUG_DISPLAY.setDebugBins(
                                    e.data[e.data.length - 5]);
                            }
                        }

                        //if debugging, update the thread info display
                        if (DEBUG) {
                            DEBUG_DISPLAY.updateThreadInfoDisplay(
                                e.data[e.data.length - 4].toPrecision(6),
                                e.data[e.data.length - 3].toPrecision(6),
                                e.data[e.data.length - 2].toPrecision(6),
                                dt);
                        }
                    }

                    //callback 1 - loaded a game object, so see if we are done 
                    //and call back to main application thread if so.
                    if (e.data[e.data.length - 1] == 1) {
                        THREAD_MANAGER._numGameObjects++;

                        if (THREAD_MANAGER._numGameObjects ==
                            TOTAL_NUM_GAME_OBJECTS) {
                            main_applicaiton_threading_callback();
                        }
                    }

                    //callback 2 - we have loaded all the game objects in thread
                    //land, so we need to marshall them once and link them to 
                    //models...
                    if (e.data[e.data.length - 1] == 2) {
                        e.data.pop(); //remove control code
                        
                        AAMODEL_MANAGER.linkMeshesToGameObjects(e.data);
                    }
                }

                this.thread.postMessage([1]); //activate this thread
                this._threadPool.push(this.thread);

            //-------------------------------------------------------------
            // if ID is 2, thread type is to manage game objects
            //-------------------------------------------------------------
            } else if (id == 2) {
                this.thread2 = new Worker(
                    "js/core_engine/worker_threads/SecondThread.js");
                this.thread2.onmessage = function (e) {
                    //callback from second thread...
                }
            }
            
            //-------------------------------------------------------------
            // mark manager as initialized
            //-------------------------------------------------------------
            this.initialized = true;

        } else {
            //-------------------------------------------------------------
            // attempted to use threads in an environment that does
            // not support workers.  Throw error and exit application
            //-------------------------------------------------------------
            console.log("AAENGINE ERROR: Browser does not support worker threads");
            console.log("AAENGINE ERROR: Calling Shutdown");
            main_exit();
        }
    }
    //---------------------------------------------------------------------


    //-----[ THREAD_MANAGER: RESIZE ]--------------------------------------
    // resize event - pass updated information to game obj manager thread
    //---------------------------------------------------------------------
    resize() {
        if (this.initialized) {
            this.thread.postMessage([2,
                GAME_WINDOW_ZOOM_EFFECT,
                GAME_WINDOW.GAME_WINDOW_SCALE,
                GAME_WINDOW.GAME_WINDOW_HALF_WIDTH,
                GAME_WINDOW.GAME_WINDOW_HALF_HEIGHT,
                GAME_WINDOW.GAME_WINDOW_UNSCALED_WIDTH,
                GAME_WINDOW.GAME_WINDOW_UNSCALED_HEIGHT,
                GAME_WINDOW.GAME_WINDOW_UNSCALED_HALF_WIDTH,
                GAME_WINDOW.GAME_WINDOW_UNSCALED_HALF_HEIGHT,
                COLLISION_BINS_X, COLLISION_BINS_Y]);
        }
    }
    //---------------------------------------------------------------------


    //-----[ THREAD_MANAGER: STOP ]----------------------------------------
    // stop event - call stop on all threads
    //---------------------------------------------------------------------
    stop() {
        for (var i = 0; i < this._threadPool.length; i++) {
            this._threadPool[i].postMessage([3]);
        }
    }
    //---------------------------------------------------------------------


    //-----[ THREAD_MANAGER: ADDGAMEOBJ ]----------------------------------
    // addGameObject - add a game object via threaded interface
    //---------------------------------------------------------------------
    addGameObject(go) {

        this.thread.postMessage([4, go]); //load up some test object...
    }
    //---------------------------------------------------------------------


    //-----[ THREAD_MANAGER: SENDINPUT ]-----------------------------------
    // sendInput - send copies of latest input inforamation to threaded
    // environment to be used by game objects
    //---------------------------------------------------------------------
    sendInput(info) {
        this.thread.postMessage([6, info]);
    }
    //---------------------------------------------------------------------


    //-----[ THREAD_MANAGER: UPDATEDEBUGCANVASSTATE ]----------------------
    // toggle state of debug canvas var inside the threaded environment
    //---------------------------------------------------------------------
    updateDebugCanvasState(bVal) {
        this.thread.postMessage([5, bVal]);
    }
    //---------------------------------------------------------------------


    //-----[ THREAD_MANAGER: LINKMODELSTOOBJECTS ]-------------------------
    // call to initiate the process of linking models to game objects
    //---------------------------------------------------------------------
    linkModelsToObjects() {
        this.thread.postMessage([7]);
    }
    //---------------------------------------------------------------------
}
//-------------------------------------------------------------------------


//-------------------------------------------------------------------------
// creation of global THREAD_MANAGER -- initialization called from
// main application init
//-------------------------------------------------------------------------
var THREAD_MANAGER = THREAD_MANAGER || new ThreadManager();
//-------------------------------------------------------------------------